@@ -10,7 +10,7 @@
 //! Configuration for `Endpoint`s.
 
 use serde::{Deserialize, Serialize};
-use std::{net::IpAddr, sync::Arc, time::Duration};
+use std::{future::Future, net::IpAddr, sync::Arc, time::Duration};
 
 /// Default for [`Config::idle_timeout`] (1 minute).
 ///
@@ -36,7 +36,7 @@ pub const DEFAULT_INITIAL_RETRY_INTERVAL: Duration = Duration::from_millis(500);
 /// gives 5-6 retries in ~30 s total retry time.
 pub const DEFAULT_MAX_RETRY_INTERVAL: Duration = Duration::from_secs(15);
 
-/// Default for [`RetryConfig::retry_interval_multiplier`] (x1.5).
+/// Default for [`RetryConfig::retry_delay_multiplier`] (x1.5).
 ///
 /// Together with the default max and initial,
 /// gives 5-6 retries in ~30 s total retry time.
@@ -140,21 +140,6 @@ pub struct Config {
     pub retry_config: RetryConfig,
 }
 
-/// Config that has passed validation.
-///
-/// Generally this is a copy of [`Config`] without optional values where we would use defaults.
-#[derive(Clone, Debug)]
-pub(crate) struct InternalConfig {
-    pub(crate) client: quinn::ClientConfig,
-    pub(crate) server: quinn::ServerConfig,
-    #[cfg(feature = "igd")]
-    pub(crate) forward_port: bool,
-    pub(crate) external_port: Option<u16>,
-    pub(crate) external_ip: Option<IpAddr>,
-    pub(crate) upnp_lease_duration: Duration,
-    pub(crate) retry_config: RetryConfig,
-}
-
 /// Retry configurations for establishing connections and sending messages.
 /// Determines the retry behaviour of requests, by setting the back off strategy used.
 #[derive(Clone, Debug, Copy, Serialize, Deserialize)]
@@ -187,6 +172,29 @@ pub struct RetryConfig {
     pub retrying_max_elapsed_time: Duration,
 }
 
+impl RetryConfig {
+    // Perform `op` and retry on errors as specified by this configuration.
+    //
+    // Note that `backoff::Error<E>` implements `From<E>` for any `E` by creating a
+    // `backoff::Error::Transient`, meaning that errors will be retried unless explicitly returning
+    // `backoff::Error::Permanent`.
+    pub(crate) fn retry<R, E, Fn, Fut>(&self, op: Fn) -> impl Future<Output = Result<R, E>>
+    where
+        Fn: FnMut() -> Fut,
+        Fut: Future<Output = Result<R, backoff::Error<E>>>,
+    {
+        let backoff = backoff::ExponentialBackoff {
+            initial_interval: self.initial_retry_interval,
+            randomization_factor: self.retry_delay_rand_factor,
+            multiplier: self.retry_delay_multiplier,
+            max_interval: self.max_retry_interval,
+            max_elapsed_time: Some(self.retrying_max_elapsed_time),
+            ..Default::default()
+        };
+        backoff::future::retry(backoff, op)
+    }
+}
+
 impl Default for RetryConfig {
     fn default() -> Self {
         Self {
@@ -199,6 +207,21 @@ impl Default for RetryConfig {
     }
 }
 
+/// Config that has passed validation.
+///
+/// Generally this is a copy of [`Config`] without optional values where we would use defaults.
+#[derive(Clone, Debug)]
+pub(crate) struct InternalConfig {
+    pub(crate) client: quinn::ClientConfig,
+    pub(crate) server: quinn::ServerConfig,
+    #[cfg(feature = "igd")]
+    pub(crate) forward_port: bool,
+    pub(crate) external_port: Option<u16>,
+    pub(crate) external_ip: Option<IpAddr>,
+    pub(crate) upnp_lease_duration: Duration,
+    pub(crate) retry_config: Arc<RetryConfig>,
+}
+
 impl InternalConfig {
     pub(crate) fn try_from_config(config: Config) -> Result<Self> {
         let idle_timeout = config.idle_timeout.unwrap_or(DEFAULT_IDLE_TIMEOUT);
@@ -221,7 +244,7 @@ impl InternalConfig {
             external_port: config.external_port,
             external_ip: config.external_ip,
             upnp_lease_duration,
-            retry_config: config.retry_config,
+            retry_config: Arc::new(config.retry_config),
         })
     }
 
