@@ -13,7 +13,7 @@ use super::wire_msg::WireMsg;
 use super::{
     config::{Config, InternalConfig, RetryConfig},
     connection_deduplicator::{ConnectionDeduplicator, DedupHandle},
-    connection_pool::{ConnId, ConnectionPool},
+    connection_pool::{ConnId, ConnectionPool, ConnectionRemover},
     connections::{
         listen_for_incoming_connections, listen_for_incoming_messages, Connection,
         DisconnectionEvents, RecvStream, SendStream,
@@ -23,7 +23,6 @@ use super::{
         SerializationError,
     },
 };
-use backoff::{future::retry, ExponentialBackoff};
 use bytes::Bytes;
 use std::net::SocketAddr;
 use tokio::sync::broadcast::{self, Sender};
@@ -67,8 +66,7 @@ impl IncomingConnections {
     }
 }
 
-/// Endpoint instance which can be used to create connections to peers,
-/// and listen to incoming messages from other peers.
+/// Endpoint instance which can be used to communicate with peers.
 #[derive(Clone)]
 pub struct Endpoint<I: ConnId> {
     local_addr: SocketAddr,
@@ -127,7 +125,7 @@ impl<I: ConnId> Endpoint<I> {
             IncomingConnections,
             IncomingMessages,
             DisconnectionEvents,
-            Option<SocketAddr>,
+            Option<Connection<I>>,
         ),
         EndpointError,
     > {
@@ -140,7 +138,7 @@ impl<I: ConnId> Endpoint<I> {
             Self::build_endpoint(local_addr.into(), config, builder)?;
 
         let contact = endpoint.connect_to_any(contacts).await;
-        let public_addr = endpoint.resolve_public_addr(contact).await?;
+        let public_addr = endpoint.resolve_public_addr(contact.as_ref()).await?;
 
         #[cfg(feature = "igd")]
         if endpoint.config.forward_port {
@@ -166,11 +164,14 @@ impl<I: ConnId> Endpoint<I> {
             endpoint.clone(),
         );
 
-        if let Some(peer) = contact {
+        if let Some(contact) = contact.as_ref() {
             let valid = endpoint
-                .endpoint_verification(peer, public_addr)
+                .endpoint_verification(contact, public_addr)
                 .await
-                .map_err(|error| EndpointError::EndpointVerification { peer, error })?;
+                .map_err(|error| EndpointError::EndpointVerification {
+                    peer: contact.remote_address(),
+                    error,
+                })?;
             if !valid {
                 return Err(EndpointError::Unreachable { public_addr });
             }
@@ -255,50 +256,32 @@ impl<I: ConnId> Endpoint<I> {
             });
     }
 
-    /// Connects to another peer, retries for `config.retry_duration_msec` if the connection fails.
+    /// Connect to a peer.
+    ///
+    /// Atttempts to connect to a peer at the given address. Connection attempts are retried based
+    /// on the [`Config::retry_config`] used to create the endpoint.
+    ///
+    /// Returns a [`Connection`], which is a handle representing the underlying connection. There's
+    /// presently little that you can do with a `Connection` itself, besides obtaining the
+    /// [`remote_address`](Connection::remote_address) which can then be used with other `Endpoint`
+    /// methods to communicate with the peer (the connection is retrieved via the connection pool –
+    /// see below).
     ///
     /// **Note:** this method is intended for use when it's necessary to connect to a specific peer.
     /// See [`connect_to_any`](Self::connect_to_any) if you just need a connection with any of a set
     /// of peers.
     ///
-    /// Returns `Connection` which is a handle for sending messages to the peer and
-    /// `IncomingMessages` which is a stream of messages received from the peer.
-    /// The incoming messages stream might be `None`. See the next section for more info.
-    ///
     /// # Connection pooling
     ///
-    /// Connection are stored in an internal pool and reused if possible. A connection remains in
-    /// the pool while its `IncomingMessages` instances exists and while the connection is open.
-    ///
-    /// When a new connection is established, this function returns both the `Connection` instance
-    /// and the `IncomingMessages` stream. If an existing connection is retrieved from the pool,
-    /// the incoming messages will be `None`. Multiple `Connection` instances can exists
-    /// simultaneously and they all share the same underlying connection resource. On the other
-    /// hand, at most one `IncomingMessages` stream can exist per peer.
-    ///
-    /// How to handle the `IncomingMessages` depends on the networking model of the application:
-    ///
-    /// In the peer-to-peer model, where peers can arbitrarily send and receive messages to/from
-    /// other peers, it is recommended to keep the `IncomingMessages` around and listen on it for
-    /// new messages by repeatedly calling `next` and only drop it when it returns `None`.
-    /// On the other hand, there is no need to keep `Connection` around as it can be cheaply
-    /// retrieved again when needed by calling `connect_to`. When the connection gets closed by the
-    /// peer or it timeouts due to inactivity, the incoming messages stream gets closed and once
-    /// it's dropped the connection gets removed from the pool automatically. Calling `connect_to`
-    /// afterwards will open a new connection.
-    ///
-    /// In the client-server model, where only the client send requests to the server and then
-    /// listens for responses and never the other way around, it's OK to ignore (drop) the incoming
-    /// messages stream and only use bi-directional streams obtained by calling
-    /// `Connection::open_bi`. In this case the connection won't be pooled and the application is
-    /// responsible for caching it.
-    ///
-    /// When sending a message on `Connection` fails, the connection is also automatically removed
-    /// from the pool and the subsequent call to `connect_to` is guaranteed to reopen new connection
-    /// too.
-    pub async fn connect_to(&self, node_addr: &SocketAddr) -> Result<(), ConnectionError> {
-        let _ = self.get_or_connect_to(node_addr).await?;
-        Ok(())
+    /// Connections are stored in an internal pool and reused if possible. A connection remains in
+    /// the pool until either side closes the connection (including due to timeouts or errors). This
+    /// method will check the pool before opening a new connection. If a new connection is opened,
+    /// it will be added to the pool.
+    pub async fn connect_to(
+        &self,
+        node_addr: &SocketAddr,
+    ) -> Result<Connection<I>, ConnectionError> {
+        self.get_or_connect_to(node_addr).await
     }
 
     /// Connect to any of the given peers.
@@ -307,11 +290,16 @@ impl<I: ConnId> Endpoint<I> {
     /// rather than having to connect to specific nodes. This method will start connecting to every
     /// peer in `peer_addrs`, and return the address of the first successfully established
     /// connection (the rest are cancelled and discarded). All connection attempts will be retried
-    /// for `config.retry_duration_msec` on failure.
+    /// based on the [`Config::retry_config`] used to create the endpoint.
     ///
-    /// The successful connection, if any, will be stored in the connection pool (see
-    /// [`connect_to`](Self::connect_to) for more info on connection pooling).
-    pub async fn connect_to_any(&self, peer_addrs: &[SocketAddr]) -> Option<SocketAddr> {
+    /// # Connection pooling
+    ///
+    /// Connections are stored in an internal pool and reused if possible. A connection remains in
+    /// the pool until either side closes the connection (including due to timeouts or errors). This
+    /// method will check the pool before opening each connection. If a new connection is opened, it
+    /// will be added to the pool. Note that already pooled connections will have a higher chance of
+    /// 'winning' the race, and being the selected peer.
+    pub async fn connect_to_any(&self, peer_addrs: &[SocketAddr]) -> Option<Connection<I>> {
         trace!("Connecting to any of {:?}", peer_addrs);
         if peer_addrs.is_empty() {
             return None;
@@ -323,7 +311,7 @@ impl<I: ConnId> Endpoint<I> {
             .map(|addr| Box::pin(self.get_or_connect_to(addr)));
 
         match futures::future::select_ok(tasks).await {
-            Ok((connection, _)) => Some(connection.remote_address()),
+            Ok((connection, _)) => Some(connection),
             Err(error) => {
                 error!("Failed to bootstrap to the network, last error: {}", error);
                 None
@@ -334,6 +322,13 @@ impl<I: ConnId> Endpoint<I> {
     /// Verify if an address is publicly reachable. This will attempt to create
     /// a new connection and use it to exchange a message and verify that the node
     /// can be reached.
+    ///
+    /// # Connection pooling
+    ///
+    /// Note that unlike most methods on `Endpoint`, this **will not** use the connection pool. This
+    /// ensures that the reachability check is accurate, otherwise we may use a pooled connection
+    /// that was opened by the peer (which tells us nothing about whether we can open a connection
+    /// to the peer).
     pub async fn is_reachable(&self, peer_addr: &SocketAddr) -> Result<(), RpcError> {
         trace!("Checking is reachable");
 
@@ -364,24 +359,39 @@ impl<I: ConnId> Endpoint<I> {
         }
     }
 
-    /// Get the connection ID of an existing connection with the provided socket address
-    pub async fn get_connection_id(&self, addr: &SocketAddr) -> Option<I> {
+    /// Get the existing `Connection` for a `SocketAddr`.
+    pub async fn get_connection_by_addr(&self, addr: &SocketAddr) -> Option<Connection<I>> {
         self.connection_pool
             .get_by_addr(addr)
             .await
-            .map(|(_, remover)| remover.id())
+            .map(|(connection, remover)| self.wrap_connection(connection, remover))
     }
 
-    /// Get the SocketAddr of a connection using the connection ID
-    pub async fn get_socket_addr_by_id(&self, addr: &I) -> Option<SocketAddr> {
+    /// Get the existing `Connection` for the given ID.
+    pub async fn get_connection_by_id(&self, id: &I) -> Option<Connection<I>> {
         self.connection_pool
-            .get_by_id(addr)
+            .get_by_id(id)
             .await
-            .map(|(_, remover)| *remover.remote_addr())
+            .map(|(connection, remover)| self.wrap_connection(connection, remover))
     }
 
-    /// Open a bi-directional peer with a given peer
-    /// Priority default is 0. Both lower and higher can be passed in.
+    /// Open a bi-directional stream with a given peer.
+    ///
+    /// # Priority
+    ///
+    /// Locally buffered data from streams with higher priority will be transmitted before data from
+    /// streams with lower priority. Changing the priority of a stream with pending data may only
+    /// take effect after that data has been transmitted. Using many different priority levels per
+    /// connection may have a negative impact on performance.
+    ///
+    /// `0` is a sensible default for 'normal' priority.
+    ///
+    /// # Connection pooling
+    ///
+    /// Connections are stored in an internal pool and reused if possible. A connection remains in
+    /// the pool until either side closes the connection (including due to timeouts or errors). This
+    /// method will check the pool before opening a new connection. If a new connection is opened,
+    /// it will be added to the pool.
     pub async fn open_bidirectional_stream(
         &self,
         peer_addr: &SocketAddr,
@@ -393,7 +403,20 @@ impl<I: ConnId> Endpoint<I> {
 
     /// Send a message to a peer over an existing connection.
     ///
-    /// Priority default is 0. Both lower and higher can be passed in.
+    /// # Priority
+    ///
+    /// Locally buffered data from streams with higher priority will be transmitted before data from
+    /// streams with lower priority. Changing the priority of a stream with pending data may only
+    /// take effect after that data has been transmitted. Using many different priority levels per
+    /// connection may have a negative impact on performance.
+    ///
+    /// `0` is a sensible default for 'normal' priority.
+    ///
+    /// # Connection pooling
+    ///
+    /// Note that unlike most methods on `Endpoint`, this **will not** use the connection pool. This
+    /// method is intended to be used when it's necessary to send a message on an existing
+    /// connection only.
     ///
     /// # Errors
     ///
@@ -409,11 +432,19 @@ impl<I: ConnId> Endpoint<I> {
         self.try_send_message_with(msg, dest, priority, None).await
     }
 
-    /// Send a message to a peer over an existing connection.
+    /// Send a message to a peer over an existing connection using given retry configuration.
     ///
-    /// Priority default is 0. Both lower and higher can be passed in.
+    /// The given `retries`, if any, will override the [`Config::retry_config`] used to create the
+    /// endpoint.
     ///
-    /// Optional retry config is passed in, for custom retry policy.
+    /// # Priority
+    ///
+    /// Locally buffered data from streams with higher priority will be transmitted before data from
+    /// streams with lower priority. Changing the priority of a stream with pending data may only
+    /// take effect after that data has been transmitted. Using many different priority levels per
+    /// connection may have a negative impact on performance.
+    ///
+    /// `0` is a sensible default for 'normal' priority.
     ///
     /// # Errors
     ///
@@ -425,57 +456,21 @@ impl<I: ConnId> Endpoint<I> {
         msg: Bytes,
         dest: &SocketAddr,
         priority: i32,
-        retries: Option<RetryConfig>,
+        retries: Option<&RetryConfig>,
     ) -> Result<(), Option<SendError>> {
         if let Some((conn, guard)) = self.connection_pool.get_by_addr(dest).await {
             trace!("Connection exists in the connection pool: {}", dest);
-            let connection = Connection::new(conn, guard);
-            let msg_retry_cfg = retries.unwrap_or(self.config.retry_config);
-            Self::retry(msg_retry_cfg, || async {
-                Ok(connection.send_uni(msg.clone(), priority).await?)
-            })
-            .await?;
+            let connection = self.wrap_connection(conn, guard);
+            retries
+                .unwrap_or(&self.config.retry_config)
+                .retry(|| async { Ok(connection.send_uni(msg.clone(), priority).await?) })
+                .await?;
             Ok(())
         } else {
             Err(None)
         }
     }
 
-    /// Sends a message to a peer. This will attempt to use an existing connection
-    /// to the peer first. If this connection is broken or doesn't exist
-    /// a new connection is created and the message is sent.
-    /// Priority default is 0. Both lower and higher can be passed in.
-    pub async fn send_message(
-        &self,
-        msg: Bytes,
-        dest: &SocketAddr,
-        priority: i32,
-    ) -> Result<(), SendError> {
-        self.send_message_with(msg, dest, priority, None).await
-    }
-
-    /// Sends a message to a peer. This will attempt to use an existing connection
-    /// to the peer first. If this connection is broken or doesn't exist
-    /// a new connection is created and the message is sent.
-    /// Priority default is 0. Both lower and higher can be passed in.
-    /// Optional retry config is passed in, for custom retry policy.
-    pub async fn send_message_with(
-        &self,
-        msg: Bytes,
-        dest: &SocketAddr,
-        priority: i32,
-        retries: Option<RetryConfig>,
-    ) -> Result<(), SendError> {
-        let connection = self.get_or_connect_to(dest).await?;
-        let msg_retry_cfg = retries.unwrap_or(self.config.retry_config);
-        Self::retry(msg_retry_cfg, || async {
-            Ok(connection.send_uni(msg.clone(), priority).await?)
-        })
-        .await?;
-
-        Ok(())
-    }
-
     /// Close all the connections of this endpoint immediately and stop accepting new connections.
     pub fn close(&self) {
         let _ = self.termination_tx.send(());
@@ -490,7 +485,7 @@ impl<I: ConnId> Endpoint<I> {
         let completion = loop {
             if let Some((conn, remover)) = self.connection_pool.get_by_addr(addr).await {
                 trace!("We are already connected to this peer: {}", addr);
-                return Ok(Connection::new(conn, remover));
+                return Ok(self.wrap_connection(conn, remover));
             }
 
             // Check if a connect attempt to this address is already in progress.
@@ -522,7 +517,7 @@ impl<I: ConnId> Endpoint<I> {
                 );
 
                 let _ = completion.complete(Ok(()));
-                Ok(Connection::new(connection, remover))
+                Ok(self.wrap_connection(connection, remover))
             }
             Err(error) => {
                 let _ = completion.complete(Err(error.clone()));
@@ -539,50 +534,53 @@ impl<I: ConnId> Endpoint<I> {
         &self,
         node_addr: &SocketAddr,
     ) -> Result<quinn::NewConnection, ConnectionError> {
-        Self::retry(self.config.retry_config, || async {
-            trace!("Attempting to connect to {:?}", node_addr);
-            let connecting = match self.quic_endpoint.connect_with(
-                self.config.client.clone(),
-                node_addr,
-                CERT_SERVER_NAME,
-            ) {
-                Ok(conn) => Ok(conn),
-                Err(error) => {
-                    warn!("Connection attempt failed due to {:?}", error);
-                    Err(ConnectionError::from(error))
-                }
-            }?;
+        self.config
+            .retry_config
+            .retry(|| async {
+                trace!("Attempting to connect to {:?}", node_addr);
+                let connecting = match self.quic_endpoint.connect_with(
+                    self.config.client.clone(),
+                    node_addr,
+                    CERT_SERVER_NAME,
+                ) {
+                    Ok(conn) => Ok(conn),
+                    Err(error) => {
+                        warn!("Connection attempt failed due to {:?}", error);
+                        Err(ConnectionError::from(error))
+                    }
+                }?;
 
-            let new_conn = match connecting.await {
-                Ok(new_conn) => {
-                    debug!("okay was had");
-                    Ok(new_conn)
-                }
-                Err(error) => {
-                    error!("some error: {:?}", error);
-                    Err(ConnectionError::from(error))
-                }
-            }?;
+                let new_conn = match connecting.await {
+                    Ok(new_conn) => {
+                        debug!("okay was had");
+                        Ok(new_conn)
+                    }
+                    Err(error) => {
+                        error!("some error: {:?}", error);
+                        Err(ConnectionError::from(error))
+                    }
+                }?;
 
-            Ok(new_conn)
-        })
-        .await
+                Ok(new_conn)
+            })
+            .await
     }
 
     // set an appropriate public address based on `config` and a reachability check.
     async fn resolve_public_addr(
         &mut self,
-        contact: Option<SocketAddr>,
+        contact: Option<&Connection<I>>,
     ) -> Result<SocketAddr, EndpointError> {
         let mut public_addr = self.local_addr;
 
         // get the IP seen for us by our contact
-        let visible_addr = if let Some(peer) = contact {
-            Some(
-                self.endpoint_echo(peer)
-                    .await
-                    .map_err(|error| EndpointError::EndpointEcho { peer, error })?,
-            )
+        let visible_addr = if let Some(contact) = contact {
+            Some(self.endpoint_echo(contact).await.map_err(|error| {
+                EndpointError::EndpointEcho {
+                    peer: contact.remote_address(),
+                    error,
+                }
+            })?)
         } else {
             None
         };
@@ -643,9 +641,9 @@ impl<I: ConnId> Endpoint<I> {
         Ok(public_addr)
     }
 
-    /// Perform the endpoint echo RPC with the given peer.
-    async fn endpoint_echo(&self, peer: SocketAddr) -> Result<SocketAddr, RpcError> {
-        let (mut send, mut recv) = self.open_bidirectional_stream(&peer, 0).await?;
+    /// Perform the endpoint echo RPC with the given contact.
+    async fn endpoint_echo(&self, contact: &Connection<I>) -> Result<SocketAddr, RpcError> {
+        let (mut send, mut recv) = contact.open_bi(0).await?;
 
         send.send(WireMsg::EndpointEchoReq).await?;
 
@@ -658,10 +656,10 @@ impl<I: ConnId> Endpoint<I> {
     /// Perform the endpoint verification RPC with the given peer.
     async fn endpoint_verification(
         &self,
-        peer: SocketAddr,
+        contact: &Connection<I>,
         public_addr: SocketAddr,
     ) -> Result<bool, RpcError> {
-        let (mut send, mut recv) = self.open_bidirectional_stream(&peer, 0).await?;
+        let (mut send, mut recv) = contact.open_bi(0).await?;
 
         send.send(WireMsg::EndpointVerificationReq(public_addr))
             .await?;
@@ -672,20 +670,13 @@ impl<I: ConnId> Endpoint<I> {
         }
     }
 
-    fn retry<R, E, Fn, Fut>(cfg: RetryConfig, op: Fn) -> impl futures::Future<Output = Result<R, E>>
-    where
-        Fn: FnMut() -> Fut,
-        Fut: futures::Future<Output = Result<R, backoff::Error<E>>>,
-    {
-        let backoff = ExponentialBackoff {
-            initial_interval: cfg.initial_retry_interval,
-            randomization_factor: cfg.retry_delay_rand_factor,
-            multiplier: cfg.retry_delay_multiplier,
-            max_interval: cfg.max_retry_interval,
-            max_elapsed_time: Some(cfg.retrying_max_elapsed_time),
-            ..Default::default()
-        };
-        retry(backoff, op)
+    /// Wrap a quinn connection, setting the default retry config and pool remover.
+    fn wrap_connection(
+        &self,
+        connection: quinn::Connection,
+        remover: ConnectionRemover<I>,
+    ) -> Connection<I> {
+        Connection::new(connection, self.config.retry_config.clone(), remover)
     }
 }
 
